@@ -12,23 +12,35 @@
 
 #include "postgres.h"
 #include "funcapi.h"
+#include "libpq-fe.h"
 #include "miscadmin.h"
 
 #include "access/htup_details.h"
+#include "access/sysattr.h"
 #include "access/xact.h"
+#include "catalog/dependency.h"
 #include "catalog/indexing.h"
+#include "catalog/pg_constraint.h"
+#include "catalog/pg_extension.h"
+#include "catalog/pg_namespace.h"
 #include "catalog/pg_type.h"
+#include "commands/extension.h"
+#include "distributed/connection_management.h"
 #include "distributed/citus_nodes.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/multi_join_order.h"
 #include "distributed/multi_logical_optimizer.h"
+#include "distributed/pg_dist_colocation.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
 #include "distributed/pg_dist_shard_placement.h"
 #include "distributed/relay_utility.h"
+#include "distributed/resource_lock.h"
+#include "distributed/remote_commands.h"
 #include "distributed/worker_manager.h"
+#include "distributed/worker_protocol.h"
 #include "nodes/makefuncs.h"
 #include "parser/scansup.h"
 #include "storage/lmgr.h"
@@ -45,11 +57,293 @@
 
 /* Local functions forward declarations */
 static uint64 * AllocateUint64(uint64 value);
+static void RecordDistributedRelationDependencies(Oid distributedRelationId,
+												  Node *distributionKey);
+static ShardPlacement * TupleToShardPlacement(TupleDesc tupleDesc,
+											  HeapTuple heapTuple);
+static uint64 DistributedTableSize(Oid relationId, char *sizeQuery);
+static uint64 DistributedTableSizeOnWorker(WorkerNode *workerNode, Oid relationId,
+										   char *sizeQuery);
+static List * ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId);
+static StringInfo GenerateSizeQueryOnMultiplePlacements(Oid distributedRelationId,
+														List *shardIntervalList,
+														char *sizeQuery);
+static void ErrorIfNotSuitableToGetSize(Oid relationId);
 
 
 /* exports for SQL callable functions */
-PG_FUNCTION_INFO_V1(master_stage_shard_row);
-PG_FUNCTION_INFO_V1(master_stage_shard_placement_row);
+PG_FUNCTION_INFO_V1(citus_table_size);
+PG_FUNCTION_INFO_V1(citus_total_relation_size);
+PG_FUNCTION_INFO_V1(citus_relation_size);
+
+
+/*
+ * citus_total_relation_size accepts a table name and returns a distributed table
+ * and its indexes' total relation size.
+ */
+Datum
+citus_total_relation_size(PG_FUNCTION_ARGS)
+{
+	Oid relationId = PG_GETARG_OID(0);
+	uint64 totalRelationSize = 0;
+
+	CheckCitusVersion(ERROR);
+
+	totalRelationSize = DistributedTableSize(relationId,
+											 PG_TOTAL_RELATION_SIZE_FUNCTION);
+
+	PG_RETURN_INT64(totalRelationSize);
+}
+
+
+/*
+ * citus_table_size accepts a table name and returns a distributed table's total
+ * relation size.
+ */
+Datum
+citus_table_size(PG_FUNCTION_ARGS)
+{
+	Oid relationId = PG_GETARG_OID(0);
+	uint64 tableSize = 0;
+
+	CheckCitusVersion(ERROR);
+
+	tableSize = DistributedTableSize(relationId, PG_TABLE_SIZE_FUNCTION);
+
+	PG_RETURN_INT64(tableSize);
+}
+
+
+/*
+ * citus_relation_size accept a table name and returns a relation's 'main'
+ * fork's size.
+ */
+Datum
+citus_relation_size(PG_FUNCTION_ARGS)
+{
+	Oid relationId = PG_GETARG_OID(0);
+	uint64 relationSize = 0;
+
+	CheckCitusVersion(ERROR);
+
+	relationSize = DistributedTableSize(relationId, PG_RELATION_SIZE_FUNCTION);
+
+	PG_RETURN_INT64(relationSize);
+}
+
+
+/*
+ * DistributedTableSize is helper function for each kind of citus size functions.
+ * It first checks whether the table is distributed and size query can be run on
+ * it. Connection to each node has to be established to get the size of the table.
+ */
+static uint64
+DistributedTableSize(Oid relationId, char *sizeQuery)
+{
+	Relation relation = NULL;
+	Relation pgDistNode = NULL;
+	List *workerNodeList = NULL;
+	ListCell *workerNodeCell = NULL;
+	uint64 totalRelationSize = 0;
+
+	if (XactModificationLevel == XACT_MODIFICATION_MULTI_SHARD)
+	{
+		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
+						errmsg("citus size functions cannot be called in transaction"
+							   " blocks which contain multi-shard data modifications")));
+	}
+
+	/* try to open relation, will error out if the relation does not exist */
+	relation = relation_open(relationId, AccessShareLock);
+
+	ErrorIfNotSuitableToGetSize(relationId);
+
+	pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
+
+	workerNodeList = ActiveWorkerNodeList();
+
+	foreach(workerNodeCell, workerNodeList)
+	{
+		WorkerNode *workerNode = (WorkerNode *) lfirst(workerNodeCell);
+		uint64 relationSizeOnNode = DistributedTableSizeOnWorker(workerNode, relationId,
+																 sizeQuery);
+		totalRelationSize += relationSizeOnNode;
+	}
+
+	heap_close(pgDistNode, NoLock);
+	heap_close(relation, AccessShareLock);
+
+	return totalRelationSize;
+}
+
+
+/*
+ * DistributedTableSizeOnWorker gets the workerNode and relationId to calculate
+ * size of that relation on the given workerNode by summing up the size of each
+ * shard placement.
+ */
+static uint64
+DistributedTableSizeOnWorker(WorkerNode *workerNode, Oid relationId, char *sizeQuery)
+{
+	StringInfo tableSizeQuery = NULL;
+	StringInfo tableSizeStringInfo = NULL;
+	char *workerNodeName = workerNode->workerName;
+	uint32 workerNodePort = workerNode->workerPort;
+	char *tableSizeString;
+	List *sizeList = NIL;
+	uint64 tableSize = 0;
+
+	List *shardIntervalsOnNode = ShardIntervalsOnWorkerNode(workerNode, relationId);
+
+	tableSizeQuery = GenerateSizeQueryOnMultiplePlacements(relationId,
+														   shardIntervalsOnNode,
+														   sizeQuery);
+
+	sizeList = ExecuteRemoteQuery(workerNodeName, workerNodePort, NULL, tableSizeQuery);
+
+	if (sizeList == NIL)
+	{
+		ereport(ERROR, (errcode(ERRCODE_CONNECTION_FAILURE),
+						errmsg("cannot get the size because of a connection error")));
+	}
+
+	tableSizeStringInfo = (StringInfo) linitial(sizeList);
+	tableSizeString = tableSizeStringInfo->data;
+	tableSize = atol(tableSizeString);
+
+	return tableSize;
+}
+
+
+/*
+ * ShardIntervalsOnNode takes a WorkerNode then compares it with each placement
+ * of table. It returns shard intervals of table on that node as a list of shard
+ * intervals. Note that, shard intervals returned as elements of the list are
+ * not the copies but the pointers.
+ *
+ * DO NOT modify the shard intervals returned by this function.
+ */
+static List *
+ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId)
+{
+	DistTableCacheEntry *distTableCacheEntry = DistributedTableCacheEntry(relationId);
+	char *workerNodeName = workerNode->workerName;
+	uint32 workerNodePort = workerNode->workerPort;
+	List *shardIntervalList = NIL;
+	int shardIndex = 0;
+	int shardIntervalArrayLength = distTableCacheEntry->shardIntervalArrayLength;
+
+	for (shardIndex = 0; shardIndex < shardIntervalArrayLength; shardIndex++)
+	{
+		ShardPlacement *placementArray =
+			distTableCacheEntry->arrayOfPlacementArrays[shardIndex];
+		int numberOfPlacements =
+			distTableCacheEntry->arrayOfPlacementArrayLengths[shardIndex];
+		int placementIndex = 0;
+
+		for (placementIndex = 0; placementIndex < numberOfPlacements; placementIndex++)
+		{
+			ShardPlacement *placement = &placementArray[placementIndex];
+			char *shardNodeName = placement->nodeName;
+			uint32 shardNodePort = placement->nodePort;
+			uint64 shardId = placement->shardId;
+			bool metadataLock = false;
+
+			metadataLock = TryLockShardDistributionMetadata(shardId, ShareLock);
+
+			/* if the lock is not acquired warn the user */
+			if (metadataLock == false)
+			{
+				ereport(WARNING, (errcode(ERRCODE_LOCK_NOT_AVAILABLE),
+								  errmsg("lock is not acquired, size of shard %ld "
+										 "will be ignored", shardId)));
+				continue;
+			}
+
+			if (strcmp(shardNodeName, workerNodeName) == 0 &&
+				shardNodePort == workerNodePort)
+			{
+				ShardInterval *shardInterval =
+					distTableCacheEntry->sortedShardIntervalArray[shardIndex];
+				shardIntervalList = lappend(shardIntervalList, shardInterval);
+			}
+		}
+	}
+
+	return shardIntervalList;
+}
+
+
+/*
+ * GenerateSizeQueryOnMultiplePlacements generates a select size query to get
+ * size of multiple tables from the relation with distributedRelationId. Note
+ * that, different size functions supported by PG are also supported by this
+ * function changing the size query given as the last parameter to function.
+ * Format of sizeQuery is pg_*_size(%s). Examples of it can be found in the
+ * master_protocol.h
+ */
+static StringInfo
+GenerateSizeQueryOnMultiplePlacements(Oid distributedRelationId, List *shardIntervalList,
+									  char *sizeQuery)
+{
+	Oid schemaId = get_rel_namespace(distributedRelationId);
+	char *schemaName = get_namespace_name(schemaId);
+
+	StringInfo selectQuery = makeStringInfo();
+	ListCell *shardIntervalCell = NULL;
+
+	appendStringInfo(selectQuery, "SELECT ");
+
+	foreach(shardIntervalCell, shardIntervalList)
+	{
+		ShardInterval *shardInterval = (ShardInterval *) lfirst(shardIntervalCell);
+		uint64 shardId = shardInterval->shardId;
+		char *shardName = get_rel_name(distributedRelationId);
+		char *shardQualifiedName = NULL;
+		char *quotedShardName = NULL;
+		AppendShardIdToName(&shardName, shardId);
+
+		shardQualifiedName = quote_qualified_identifier(schemaName, shardName);
+		quotedShardName = quote_literal_cstr(shardQualifiedName);
+
+		appendStringInfo(selectQuery, sizeQuery, quotedShardName);
+		appendStringInfo(selectQuery, " + ");
+	}
+
+	/*
+	 * Add 0 as a last size, it handles empty list case and makes size control checks
+	 * unnecessary which would have implemented without this line.
+	 */
+	appendStringInfo(selectQuery, "0;");
+
+	return selectQuery;
+}
+
+
+/*
+ * ErrorIfNotSuitableToGetSize determines whether the table is suitable to find
+ * its' size with internal functions.
+ */
+static void
+ErrorIfNotSuitableToGetSize(Oid relationId)
+{
+	if (!IsDistributedTable(relationId))
+	{
+		char *relationName = get_rel_name(relationId);
+		char *escapedQueryString = quote_literal_cstr(relationName);
+		ereport(ERROR, (errcode(ERRCODE_INVALID_TABLE_DEFINITION),
+						errmsg("cannot calculate the size because relation %s is not "
+							   "distributed", escapedQueryString)));
+	}
+
+	if (PartitionMethod(relationId) == DISTRIBUTE_BY_HASH &&
+		!SingleReplicatedTable(relationId))
+	{
+		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						errmsg("cannot calculate the size because replication factor "
+							   "is greater than 1")));
+	}
+}
 
 
 /*
@@ -230,6 +524,24 @@ CopyShardInterval(ShardInterval *srcInterval, ShardInterval *destInterval)
 }
 
 
+/*
+ * CopyShardPlacement copies the values of the source placement into the
+ * target placement.
+ */
+void
+CopyShardPlacement(ShardPlacement *srcPlacement, ShardPlacement *destPlacement)
+{
+	/* first copy all by-value fields */
+	memcpy(destPlacement, srcPlacement, sizeof(ShardPlacement));
+
+	/* and then the fields pointing to external values */
+	if (srcPlacement->nodeName)
+	{
+		destPlacement->nodeName = pstrdup(srcPlacement->nodeName);
+	}
+}
+
+
 /*
  * ShardLength finds shard placements for the given shardId, extracts the length
  * of a finalized shard, and returns the shard's length. This function errors
@@ -321,13 +633,47 @@ FinalizedShardPlacementList(uint64 shardId)
 
 
 /*
- * ShardPlacementList finds shard placements for the given shardId from system
- * catalogs, converts these placements to their in-memory representation, and
- * returns the converted shard placements in a new list.
+ * FinalizedShardPlacement finds a shard placement for the given shardId from
+ * system catalog, chooses a placement that is in finalized state and returns
+ * that shard placement. If this function cannot find a healthy shard placement
+ * and missingOk is set to false it errors out.
+ */
+ShardPlacement *
+FinalizedShardPlacement(uint64 shardId, bool missingOk)
+{
+	List *finalizedPlacementList = FinalizedShardPlacementList(shardId);
+	ShardPlacement *shardPlacement = NULL;
+
+	if (list_length(finalizedPlacementList) == 0)
+	{
+		if (!missingOk)
+		{
+			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+							errmsg("could not find any healthy placement for shard "
+								   UINT64_FORMAT, shardId)));
+		}
+
+		return shardPlacement;
+	}
+
+	shardPlacement = (ShardPlacement *) linitial(finalizedPlacementList);
+
+	return shardPlacement;
+}
+
+
+/*
+ * BuildShardPlacementList finds shard placements for the given shardId from
+ * system catalogs, converts these placements to their in-memory
+ * representation, and returns the converted shard placements in a new list.
+ *
+ * This probably only should be called from metadata_cache.c.  Resides here
+ * because it shares code with other routines in this file.
  */
 List *
-ShardPlacementList(uint64 shardId)
+BuildShardPlacementList(ShardInterval *shardInterval)
 {
+	int64 shardId = shardInterval->shardId;
 	List *shardPlacementList = NIL;
 	Relation pgShardPlacement = NULL;
 	SysScanDesc scanDescriptor = NULL;
@@ -359,13 +705,6 @@ ShardPlacementList(uint64 shardId)
 	systable_endscan(scanDescriptor);
 	heap_close(pgShardPlacement, AccessShareLock);
 
-	/* if no shard placements are found, warn the user */
-	if (shardPlacementList == NIL)
-	{
-		ereport(WARNING, (errmsg("could not find any shard placements for shardId "
-								 UINT64_FORMAT, shardId)));
-	}
-
 	return shardPlacementList;
 }
 
@@ -375,7 +714,7 @@ ShardPlacementList(uint64 shardId)
  * and converts this tuple to in-memory struct. The function assumes the
  * caller already has locks on the tuple, and doesn't perform any locking.
  */
-ShardPlacement *
+static ShardPlacement *
 TupleToShardPlacement(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 {
 	ShardPlacement *shardPlacement = NULL;
@@ -513,6 +852,142 @@ InsertShardPlacementRow(uint64 shardId, uint64 placementId,
 }
 
 
+/*
+ * InsertIntoPgDistPartition inserts a new tuple into pg_dist_partition.
+ */
+void
+InsertIntoPgDistPartition(Oid relationId, char distributionMethod,
+						  Var *distributionColumn, uint32 colocationId,
+						  char replicationModel)
+{
+	Relation pgDistPartition = NULL;
+	char *distributionColumnString = NULL;
+
+	HeapTuple newTuple = NULL;
+	Datum newValues[Natts_pg_dist_partition];
+	bool newNulls[Natts_pg_dist_partition];
+
+	/* open system catalog and insert new tuple */
+	pgDistPartition = heap_open(DistPartitionRelationId(), RowExclusiveLock);
+
+	/* form new tuple for pg_dist_partition */
+	memset(newValues, 0, sizeof(newValues));
+	memset(newNulls, false, sizeof(newNulls));
+
+	newValues[Anum_pg_dist_partition_logicalrelid - 1] =
+		ObjectIdGetDatum(relationId);
+	newValues[Anum_pg_dist_partition_partmethod - 1] =
+		CharGetDatum(distributionMethod);
+	newValues[Anum_pg_dist_partition_colocationid - 1] = UInt32GetDatum(colocationId);
+	newValues[Anum_pg_dist_partition_repmodel - 1] = CharGetDatum(replicationModel);
+
+	/* set partkey column to NULL for reference tables */
+	if (distributionMethod != DISTRIBUTE_BY_NONE)
+	{
+		distributionColumnString = nodeToString((Node *) distributionColumn);
+
+		newValues[Anum_pg_dist_partition_partkey - 1] =
+			CStringGetTextDatum(distributionColumnString);
+	}
+	else
+	{
+		newValues[Anum_pg_dist_partition_partkey - 1] = PointerGetDatum(NULL);
+		newNulls[Anum_pg_dist_partition_partkey - 1] = true;
+	}
+
+	newTuple = heap_form_tuple(RelationGetDescr(pgDistPartition), newValues, newNulls);
+
+	/* finally insert tuple, build index entries & register cache invalidation */
+	simple_heap_insert(pgDistPartition, newTuple);
+	CatalogUpdateIndexes(pgDistPartition, newTuple);
+	CitusInvalidateRelcacheByRelid(relationId);
+
+	RecordDistributedRelationDependencies(relationId, (Node *) distributionColumn);
+
+	CommandCounterIncrement();
+	heap_close(pgDistPartition, NoLock);
+}
+
+
+/*
+ * RecordDistributedRelationDependencies creates the dependency entries
+ * necessary for a distributed relation in addition to the preexisting ones
+ * for a normal relation.
+ *
+ * We create one dependency from the (now distributed) relation to the citus
+ * extension to prevent the extension from being dropped while distributed
+ * tables exist. Furthermore a dependency from pg_dist_partition's
+ * distribution clause to the underlying columns is created, but it's marked
+ * as being owned by the relation itself. That means the entire table can be
+ * dropped, but the column itself can't. Neither can the type of the
+ * distribution column be changed (c.f. ATExecAlterColumnType).
+ */
+static void
+RecordDistributedRelationDependencies(Oid distributedRelationId, Node *distributionKey)
+{
+	ObjectAddress relationAddr = { 0, 0, 0 };
+	ObjectAddress citusExtensionAddr = { 0, 0, 0 };
+
+	relationAddr.classId = RelationRelationId;
+	relationAddr.objectId = distributedRelationId;
+	relationAddr.objectSubId = 0;
+
+	citusExtensionAddr.classId = ExtensionRelationId;
+	citusExtensionAddr.objectId = get_extension_oid("citus", false);
+	citusExtensionAddr.objectSubId = 0;
+
+	/* dependency from table entry to extension */
+	recordDependencyOn(&relationAddr, &citusExtensionAddr, DEPENDENCY_NORMAL);
+
+	/* make sure the distribution key column/expression does not just go away */
+	recordDependencyOnSingleRelExpr(&relationAddr, distributionKey, distributedRelationId,
+									DEPENDENCY_NORMAL, DEPENDENCY_NORMAL);
+}
+
+
+/*
+ * DeletePartitionRow removes the row from pg_dist_partition where the logicalrelid
+ * field equals to distributedRelationId. Then, the function invalidates the
+ * metadata cache.
+ */
+void
+DeletePartitionRow(Oid distributedRelationId)
+{
+	Relation pgDistPartition = NULL;
+	HeapTuple heapTuple = NULL;
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 1;
+
+	pgDistPartition = heap_open(DistPartitionRelationId(), RowExclusiveLock);
+
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_partition_logicalrelid,
+				BTEqualStrategyNumber, F_OIDEQ, ObjectIdGetDatum(distributedRelationId));
+
+	scanDescriptor = systable_beginscan(pgDistPartition, InvalidOid, false, NULL,
+										scanKeyCount, scanKey);
+
+	heapTuple = systable_getnext(scanDescriptor);
+	if (!HeapTupleIsValid(heapTuple))
+	{
+		ereport(ERROR, (errmsg("could not find valid entry for partition %d",
+							   distributedRelationId)));
+	}
+
+	simple_heap_delete(pgDistPartition, &heapTuple->t_self);
+
+	systable_endscan(scanDescriptor);
+
+	/* invalidate the cache */
+	CitusInvalidateRelcacheByRelid(distributedRelationId);
+
+	/* increment the counter so that next command can see the row */
+	CommandCounterIncrement();
+
+	heap_close(pgDistPartition, RowExclusiveLock);
+}
+
+
 /*
  * DeleteShardRow opens the shard system catalog, finds the unique row that has
  * the given shardId, and deletes this row.
@@ -692,6 +1167,69 @@ UpdateShardPlacementState(uint64 placementId, char shardState)
 }
 
 
+/*
+ * UpdateColocationGroupReplicationFactor finds colocation group record for given
+ * colocationId and updates its replication factor to given replicationFactor value.
+ * Since we do not cache pg_dist_colocation table, we do not need to invalidate the
+ * cache after updating replication factor.
+ */
+void
+UpdateColocationGroupReplicationFactor(uint32 colocationId, int replicationFactor)
+{
+	Relation pgDistColocation = NULL;
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 1;
+	bool indexOK = true;
+	HeapTuple heapTuple = NULL;
+	HeapTuple newHeapTuple = NULL;
+	TupleDesc tupleDescriptor = NULL;
+
+	Datum values[Natts_pg_dist_colocation];
+	bool isnull[Natts_pg_dist_colocation];
+	bool replace[Natts_pg_dist_colocation];
+
+	/* we first search for colocation group by its colocation id */
+	pgDistColocation = heap_open(DistColocationRelationId(), RowExclusiveLock);
+	tupleDescriptor = RelationGetDescr(pgDistColocation);
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_colocation_colocationid, BTEqualStrategyNumber,
+				F_OIDEQ, ObjectIdGetDatum(colocationId));
+
+	scanDescriptor = systable_beginscan(pgDistColocation,
+										DistColocationColocationidIndexId(), indexOK,
+										NULL, scanKeyCount, scanKey);
+
+	heapTuple = systable_getnext(scanDescriptor);
+	if (!HeapTupleIsValid(heapTuple))
+	{
+		ereport(ERROR, (errcode(ERRCODE_UNDEFINED_OBJECT),
+						errmsg("could not find valid entry for colocation group "
+							   "%d", colocationId)));
+	}
+
+	/* after we find colocation group, we update it with new values */
+	memset(replace, false, sizeof(replace));
+	memset(isnull, false, sizeof(isnull));
+	memset(values, 0, sizeof(values));
+
+	values[Anum_pg_dist_colocation_replicationfactor - 1] = Int32GetDatum(
+		replicationFactor);
+	replace[Anum_pg_dist_colocation_replicationfactor - 1] = true;
+
+	newHeapTuple = heap_modify_tuple(heapTuple, tupleDescriptor, values, isnull, replace);
+	simple_heap_update(pgDistColocation, &newHeapTuple->t_self, newHeapTuple);
+
+	CatalogUpdateIndexes(pgDistColocation, newHeapTuple);
+
+	CommandCounterIncrement();
+
+	heap_freetuple(newHeapTuple);
+
+	systable_endscan(scanDescriptor);
+	heap_close(pgDistColocation, NoLock);
+}
+
+
 /*
  * Check that the current user has `mode` permissions on relationId, error out
  * if not. Superusers always have such permissions.
@@ -765,137 +1303,48 @@ TableOwner(Oid relationId)
 
 
 /*
- * master_stage_shard_row() inserts a row into pg_dist_shard, after performing
- * basic permission checks.
- *
- * TODO: This function only exists for csql's \stage, and should not otherwise
- * be used. Once \stage is removed, it'll be removed too.
+ * TableReferenced function checks whether given table is referenced by another table
+ * via foreign constraints. If it is referenced, this function returns true. To check
+ * that, this function searches given relation at pg_constraints system catalog. However
+ * since there is no index for the column we searched, this function performs sequential
+ * search, therefore call this function with caution.
  */
-Datum
-master_stage_shard_row(PG_FUNCTION_ARGS)
+bool
+TableReferenced(Oid relationId)
 {
-	Oid distributedRelationId = InvalidOid;
-	uint64 shardId = 0;
-	char storageType = 0;
-	text *shardMinValue = NULL;
-	text *shardMaxValue = NULL;
-	Relation relation;
-
-	/*
-	 * Have to check arguments for NULLness as it can't be declared STRICT
-	 * because of min/max arguments, which have to be NULLable for new shards.
-	 */
-	if (PG_ARGISNULL(0))
-	{
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("relation cannot be null")));
-	}
-	else if (PG_ARGISNULL(1))
-	{
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("shard cannot be null")));
-	}
-	else if (PG_ARGISNULL(2))
-	{
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("storage type cannot be null")));
-	}
-
-	distributedRelationId = PG_GETARG_OID(0);
-	shardId = PG_GETARG_INT64(1);
-	storageType = PG_GETARG_CHAR(2);
-
-	if (!PG_ARGISNULL(3))
-	{
-		shardMinValue = PG_GETARG_TEXT_P(3);
-	}
-
-	if (!PG_ARGISNULL(4))
-	{
-		shardMaxValue = PG_GETARG_TEXT_P(4);
-	}
-
-	relation = heap_open(distributedRelationId, RowExclusiveLock);
-
-	/*
-	 * Check permissions on relation. Note we require ACL_INSERT and not owner
-	 * rights - it'd be worse for security to require every user performing
-	 * data loads to be made a table owner - besides being more complex to set
-	 * up.
-	 */
-	EnsureTablePermissions(distributedRelationId, ACL_INSERT);
-
-	/* and finally actually insert the row */
-	InsertShardRow(distributedRelationId, shardId, storageType,
-				   shardMinValue, shardMaxValue);
-
-	heap_close(relation, NoLock);
-
-	PG_RETURN_VOID();
-}
-
-
-/*
- * master_stage_shard_placement_row() inserts a row into
- * pg_dist_shard_placment, after performing some basic checks.
- *
- * TODO: This function only exists for csql's \stage, and should not otherwise
- * be used. Once \stage is removed, it'll be removed too.
- */
-Datum
-master_stage_shard_placement_row(PG_FUNCTION_ARGS)
-{
-	uint64 shardId = PG_GETARG_INT64(0);
-	int32 shardState = PG_GETARG_INT32(1);
-	int32 shardLength = PG_GETARG_INT64(2);
-	char *nodeName = text_to_cstring(PG_GETARG_TEXT_P(3));
-	int nodePort = PG_GETARG_INT32(4);
-
-	Oid distributedRelationId = InvalidOid;
-	Relation relation = NULL;
-
-	Relation pgDistShard = heap_open(DistShardRelationId(), RowExclusiveLock);
+	Relation pgConstraint = NULL;
+	HeapTuple heapTuple = NULL;
+	SysScanDesc scanDescriptor = NULL;
 	ScanKeyData scanKey[1];
 	int scanKeyCount = 1;
-	SysScanDesc scanDescriptor = NULL;
-	HeapTuple heapTuple = NULL;
+	Oid scanIndexId = InvalidOid;
+	bool useIndex = false;
 
-	/* Lookup which table the shardid belongs to */
-	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_shardid,
-				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(shardId));
+	pgConstraint = heap_open(ConstraintRelationId, AccessShareLock);
+
+	ScanKeyInit(&scanKey[0], Anum_pg_constraint_confrelid, BTEqualStrategyNumber, F_OIDEQ,
+				relationId);
+	scanDescriptor = systable_beginscan(pgConstraint, scanIndexId, useIndex, NULL,
+										scanKeyCount, scanKey);
 
-	scanDescriptor = systable_beginscan(pgDistShard,
-										DistShardShardidIndexId(), true,
-										NULL, scanKeyCount, scanKey);
 	heapTuple = systable_getnext(scanDescriptor);
-	if (HeapTupleIsValid(heapTuple))
+	while (HeapTupleIsValid(heapTuple))
 	{
-		Form_pg_dist_shard pgDistShardForm = (Form_pg_dist_shard) GETSTRUCT(heapTuple);
-		distributedRelationId = pgDistShardForm->logicalrelid;
-	}
-	else
-	{
-		ereport(ERROR, (errmsg("could not find valid entry for shard "
-							   UINT64_FORMAT, shardId)));
+		Form_pg_constraint constraintForm = (Form_pg_constraint) GETSTRUCT(heapTuple);
+
+		if (constraintForm->contype == CONSTRAINT_FOREIGN)
+		{
+			systable_endscan(scanDescriptor);
+			heap_close(pgConstraint, NoLock);
+
+			return true;
+		}
+
+		heapTuple = systable_getnext(scanDescriptor);
 	}
+
 	systable_endscan(scanDescriptor);
+	heap_close(pgConstraint, NoLock);
 
-	relation = heap_open(distributedRelationId, RowExclusiveLock);
-
-	/*
-	 * Check permissions on relation. Note we require ACL_INSERT and not owner
-	 * rights - it'd be worse for security to require every user performing
-	 * data loads to be made a table owner - besides being more complex to set
-	 * up.
-	 */
-	EnsureTablePermissions(distributedRelationId, ACL_INSERT);
-
-	/* finally insert placement */
-	InsertShardPlacementRow(shardId, INVALID_PLACEMENT_ID, shardState, shardLength,
-							nodeName, nodePort);
-
-	heap_close(relation, NoLock);
-	heap_close(pgDistShard, NoLock);
-
-	PG_RETURN_VOID();
+	return false;
 }
