@@ -17,6 +17,7 @@
 #include "postgres.h"
 #include "funcapi.h"
 #include "miscadmin.h"
+#include "libpq-fe.h"
 
 #include "access/htup_details.h"
 #include "access/xact.h"
@@ -24,6 +25,7 @@
 #include "catalog/indexing.h"
 #include "catalog/namespace.h"
 #include "distributed/colocation_utils.h"
+#include "distributed/connection_management.h"
 #include "distributed/multi_client_executor.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/master_protocol.h"
@@ -31,7 +33,10 @@
 #include "distributed/multi_join_order.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
+#include "distributed/placement_connection.h"
+#include "distributed/remote_commands.h"
 #include "distributed/resource_lock.h"
+#include "distributed/transaction_management.h"
 #include "distributed/worker_manager.h"
 #include "distributed/worker_protocol.h"
 #include "utils/builtins.h"
@@ -44,7 +49,7 @@
 
 
 /* Local functions forward declarations */
-static bool WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId,
+static bool WorkerShardStats(ShardPlacement *placement, Oid relationId,
 							 char *shardName, uint64 *shardSize,
 							 text **shardMinValue, text **shardMaxValue);
 
@@ -66,7 +71,7 @@ master_create_empty_shard(PG_FUNCTION_ARGS)
 {
 	text *relationNameText = PG_GETARG_TEXT_P(0);
 	char *relationName = text_to_cstring(relationNameText);
-	List *workerNodeList = WorkerNodeList();
+	List *workerNodeList = NIL;
 	uint64 shardId = INVALID_SHARD_ID;
 	List *ddlEventList = NULL;
 	uint32 attemptableNodeCount = 0;
@@ -82,6 +87,12 @@ master_create_empty_shard(PG_FUNCTION_ARGS)
 	Oid relationId = ResolveRelationId(relationNameText);
 	char relationKind = get_rel_relkind(relationId);
 	char *relationOwner = TableOwner(relationId);
+	char replicationModel = REPLICATION_MODEL_INVALID;
+	bool includeSequenceDefaults = false;
+
+	CheckCitusVersion(ERROR);
+
+	workerNodeList = ActiveWorkerNodeList();
 
 	EnsureTablePermissions(relationId, ACL_INSERT);
 	CheckDistributedTable(relationId);
@@ -114,12 +125,23 @@ master_create_empty_shard(PG_FUNCTION_ARGS)
 						errdetail("We currently don't support creating shards "
 								  "on hash-partitioned tables")));
 	}
+	else if (partitionMethod == DISTRIBUTE_BY_NONE)
+	{
+		ereport(ERROR, (errmsg("relation \"%s\" is a reference table",
+							   relationName),
+						errdetail("We currently don't support creating shards "
+								  "on reference tables")));
+	}
+
+	replicationModel = TableReplicationModel(relationId);
+
+	EnsureReplicationSettings(relationId, replicationModel);
 
 	/* generate new and unique shardId from sequence */
 	shardId = GetNextShardId();
 
 	/* get table DDL commands to replay on the worker node */
-	ddlEventList = GetTableDDLEvents(relationId);
+	ddlEventList = GetTableDDLEvents(relationId, includeSequenceDefaults);
 
 	/* if enough live nodes, add an extra candidate node as backup */
 	attemptableNodeCount = ShardReplicationFactor;
@@ -195,20 +217,24 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 	char *shardTableName = NULL;
 	char *shardQualifiedName = NULL;
 	List *shardPlacementList = NIL;
-	List *succeededPlacementList = NIL;
-	List *failedPlacementList = NIL;
 	ListCell *shardPlacementCell = NULL;
-	ListCell *failedPlacementCell = NULL;
 	uint64 newShardSize = 0;
 	uint64 shardMaxSizeInBytes = 0;
 	float4 shardFillLevel = 0.0;
 	char partitionMethod = 0;
 
-	ShardInterval *shardInterval = LoadShardInterval(shardId);
-	Oid relationId = shardInterval->relationId;
-	bool cstoreTable = CStoreTable(relationId);
+	ShardInterval *shardInterval = NULL;
+	Oid relationId = InvalidOid;
+	bool cstoreTable = false;
 
-	char storageType = shardInterval->storageType;
+	char storageType = 0;
+
+	CheckCitusVersion(ERROR);
+
+	shardInterval = LoadShardInterval(shardId);
+	relationId = shardInterval->relationId;
+	cstoreTable = CStoreTable(relationId);
+	storageType = shardInterval->storageType;
 
 	EnsureTablePermissions(relationId, ACL_INSERT);
 
@@ -219,19 +245,18 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 	}
 
 	partitionMethod = PartitionMethod(relationId);
-	if (partitionMethod == DISTRIBUTE_BY_HASH)
+	if (partitionMethod == DISTRIBUTE_BY_HASH || partitionMethod == DISTRIBUTE_BY_NONE)
 	{
 		ereport(ERROR, (errmsg("cannot append to shardId " UINT64_FORMAT, shardId),
 						errdetail("We currently don't support appending to shards "
-								  "in hash-partitioned tables")));
+								  "in hash-partitioned or reference tables")));
 	}
 
-	/*
-	 * We lock on the shardId, but do not unlock. When the function returns, and
-	 * the transaction for this function commits, this lock will automatically
-	 * be released. This ensures appends to a shard happen in a serial manner.
-	 */
-	LockShardResource(shardId, AccessExclusiveLock);
+	/* ensure that the shard placement metadata does not change during the append */
+	LockShardDistributionMetadata(shardId, ShareLock);
+
+	/* serialize appends to the same shard */
+	LockShardResource(shardId, ExclusiveLock);
 
 	/* get schame name of the target shard */
 	shardSchemaOid = get_rel_namespace(relationId);
@@ -251,13 +276,16 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 						errhint("Try running master_create_empty_shard() first")));
 	}
 
+	BeginOrContinueCoordinatedTransaction();
+
 	/* issue command to append table to each shard placement */
 	foreach(shardPlacementCell, shardPlacementList)
 	{
 		ShardPlacement *shardPlacement = (ShardPlacement *) lfirst(shardPlacementCell);
-		char *workerName = shardPlacement->nodeName;
-		uint32 workerPort = shardPlacement->nodePort;
-		List *queryResultList = NIL;
+		MultiConnection *connection = GetPlacementConnection(FOR_DML, shardPlacement,
+															 NULL);
+		PGresult *queryResult = NULL;
+		int executeResult = 0;
 
 		StringInfo workerAppendQuery = makeStringInfo();
 		appendStringInfo(workerAppendQuery, WORKER_APPEND_TABLE_TO_SHARD,
@@ -265,43 +293,20 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 						 quote_literal_cstr(sourceTableName),
 						 quote_literal_cstr(sourceNodeName), sourceNodePort);
 
-		/* inserting data should be performed by the current user */
-		queryResultList = ExecuteRemoteQuery(workerName, workerPort, NULL,
-											 workerAppendQuery);
-		if (queryResultList != NIL)
+		RemoteTransactionBeginIfNecessary(connection);
+
+		executeResult = ExecuteOptionalRemoteCommand(connection, workerAppendQuery->data,
+													 &queryResult);
+		PQclear(queryResult);
+		ForgetResults(connection);
+
+		if (executeResult != 0)
 		{
-			succeededPlacementList = lappend(succeededPlacementList, shardPlacement);
-		}
-		else
-		{
-			failedPlacementList = lappend(failedPlacementList, shardPlacement);
+			MarkRemoteTransactionFailed(connection, false);
 		}
 	}
 
-	/* before updating metadata, check that we appended to at least one shard */
-	if (succeededPlacementList == NIL)
-	{
-		ereport(ERROR, (errmsg("could not append table to any shard placement")));
-	}
-
-	/* make sure we don't process cancel signals */
-	HOLD_INTERRUPTS();
-
-	/* mark shard placements that we couldn't append to as inactive */
-	foreach(failedPlacementCell, failedPlacementList)
-	{
-		ShardPlacement *placement = (ShardPlacement *) lfirst(failedPlacementCell);
-		uint64 placementId = placement->placementId;
-		char *workerName = placement->nodeName;
-		uint32 workerPort = placement->nodePort;
-
-		UpdateShardPlacementState(placementId, FILE_INACTIVE);
-
-		ereport(WARNING, (errmsg("could not append table to shard \"%s\" on node "
-								 "\"%s:%u\"", shardQualifiedName, workerName,
-								 workerPort),
-						  errdetail("Marking this shard placement as inactive")));
-	}
+	MarkFailedShardPlacements();
 
 	/* update shard statistics and get new shard size */
 	newShardSize = UpdateShardStatistics(shardId);
@@ -310,8 +315,6 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 	shardMaxSizeInBytes = (int64) ShardMaxSize * 1024L;
 	shardFillLevel = ((float4) newShardSize / (float4) shardMaxSizeInBytes);
 
-	RESUME_INTERRUPTS();
-
 	PG_RETURN_FLOAT4(shardFillLevel);
 }
 
@@ -326,6 +329,8 @@ master_update_shard_statistics(PG_FUNCTION_ARGS)
 	int64 shardId = PG_GETARG_INT64(0);
 	uint64 shardSize = 0;
 
+	CheckCitusVersion(ERROR);
+
 	shardSize = UpdateShardStatistics(shardId);
 
 	PG_RETURN_INT64(shardSize);
@@ -392,18 +397,6 @@ CreateShardPlacements(Oid relationId, int64 shardId, List *ddlEventList,
 		int shardIndex = -1; /* not used in this code path */
 		bool created = false;
 
-		/*
-		 * In this code path, we create not co-located tables. Therefore we should error
-		 * out if there is a foreign key constraint specified.
-		 */
-		if (foreignConstraintCommandList != NIL)
-		{
-			ereport(ERROR, (errmsg("could only create distributed table"),
-							errdetail("Table contains foreign key constraints. Foreign "
-									  "key constraints only supported with co-located "
-									  "tables")));
-		}
-
 		created = WorkerCreateShard(relationId, nodeName, nodePort, shardIndex,
 									shardId, newPlacementOwner, ddlEventList,
 									foreignConstraintCommandList);
@@ -508,7 +501,22 @@ WorkerCreateShard(Oid relationId, char *nodeName, uint32 nodePort,
 		referencedSchemaId = get_rel_namespace(referencedRelationId);
 		referencedSchemaName = get_namespace_name(referencedSchemaId);
 		escapedReferencedSchemaName = quote_literal_cstr(referencedSchemaName);
-		referencedShardId = ColocatedShardIdInRelation(referencedRelationId, shardIndex);
+
+		/*
+		 * In case of self referencing shards, relation itself might not be distributed
+		 * already. Therefore we cannot use ColocatedShardIdInRelation which assumes
+		 * given relation is distributed. Besides, since we know foreign key references
+		 * itself, referencedShardId is actual shardId anyway.
+		 */
+		if (relationId == referencedRelationId)
+		{
+			referencedShardId = shardId;
+		}
+		else
+		{
+			referencedShardId = ColocatedShardIdInRelation(referencedRelationId,
+														   shardIndex);
+		}
 
 		appendStringInfo(applyForeignConstraintCommand,
 						 WORKER_APPLY_INTER_SHARD_DDL_COMMAND, shardId, escapedSchemaName,
@@ -561,10 +569,8 @@ UpdateShardStatistics(int64 shardId)
 	foreach(shardPlacementCell, shardPlacementList)
 	{
 		ShardPlacement *placement = (ShardPlacement *) lfirst(shardPlacementCell);
-		char *workerName = placement->nodeName;
-		uint32 workerPort = placement->nodePort;
 
-		statsOK = WorkerShardStats(workerName, workerPort, relationId, shardQualifiedName,
+		statsOK = WorkerShardStats(placement, relationId, shardQualifiedName,
 								   &shardSize, &minValue, &maxValue);
 		if (statsOK)
 		{
@@ -627,7 +633,7 @@ UpdateShardStatistics(int64 shardId)
  * we assume have changed after new table data have been appended to the shard.
  */
 static bool
-WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardName,
+WorkerShardStats(ShardPlacement *placement, Oid relationId, char *shardName,
 				 uint64 *shardSize, text **shardMinValue, text **shardMaxValue)
 {
 	char *quotedShardName = NULL;
@@ -640,11 +646,7 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 	char *partitionColumnName = NULL;
 	StringInfo partitionValueQuery = makeStringInfo();
 
-	int32 connectionId = -1;
-	bool queryOK = false;
-	void *queryResult = NULL;
-	int rowCount = 0;
-	int columnCount = 0;
+	PGresult *queryResult = NULL;
 	const int minValueIndex = 0;
 	const int maxValueIndex = 1;
 
@@ -654,16 +656,16 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 	bool minValueIsNull = false;
 	bool maxValueIsNull = false;
 
+	int connectionFlags = 0;
+	int executeCommand = 0;
+
+	MultiConnection *connection = GetPlacementConnection(connectionFlags, placement,
+														 NULL);
+
 	*shardSize = 0;
 	*shardMinValue = NULL;
 	*shardMaxValue = NULL;
 
-	connectionId = MultiClientConnect(nodeName, nodePort, NULL, NULL);
-	if (connectionId == INVALID_CONNECTION_ID)
-	{
-		return false;
-	}
-
 	quotedShardName = quote_literal_cstr(shardName);
 
 	cstoreTable = CStoreTable(relationId);
@@ -676,18 +678,18 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 		appendStringInfo(tableSizeQuery, SHARD_TABLE_SIZE_QUERY, quotedShardName);
 	}
 
-	queryOK = MultiClientExecute(connectionId, tableSizeQuery->data,
-								 &queryResult, &rowCount, &columnCount);
-	if (!queryOK)
+	executeCommand = ExecuteOptionalRemoteCommand(connection, tableSizeQuery->data,
+												  &queryResult);
+	if (executeCommand != 0)
 	{
-		MultiClientDisconnect(connectionId);
 		return false;
 	}
 
-	tableSizeString = MultiClientGetValue(queryResult, 0, 0);
+	tableSizeString = PQgetvalue(queryResult, 0, 0);
 	if (tableSizeString == NULL)
 	{
-		MultiClientDisconnect(connectionId);
+		PQclear(queryResult);
+		ForgetResults(connection);
 		return false;
 	}
 
@@ -695,20 +697,19 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 	tableSize = strtoull(tableSizeString, &tableSizeStringEnd, 0);
 	if (errno != 0 || (*tableSizeStringEnd) != '\0')
 	{
-		MultiClientClearResult(queryResult);
-		MultiClientDisconnect(connectionId);
+		PQclear(queryResult);
+		ForgetResults(connection);
 		return false;
 	}
 
 	*shardSize = tableSize;
 
-	MultiClientClearResult(queryResult);
+	PQclear(queryResult);
+	ForgetResults(connection);
 
 	if (partitionType != DISTRIBUTE_BY_APPEND)
 	{
 		/* we don't need min/max for non-append distributed tables */
-		MultiClientDisconnect(connectionId);
-
 		return true;
 	}
 
@@ -719,28 +720,27 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 	appendStringInfo(partitionValueQuery, SHARD_RANGE_QUERY,
 					 partitionColumnName, partitionColumnName, shardName);
 
-	queryOK = MultiClientExecute(connectionId, partitionValueQuery->data,
-								 &queryResult, &rowCount, &columnCount);
-	if (!queryOK)
+	executeCommand = ExecuteOptionalRemoteCommand(connection, partitionValueQuery->data,
+												  &queryResult);
+	if (executeCommand != 0)
 	{
-		MultiClientDisconnect(connectionId);
 		return false;
 	}
 
-	minValueIsNull = MultiClientValueIsNull(queryResult, 0, minValueIndex);
-	maxValueIsNull = MultiClientValueIsNull(queryResult, 0, maxValueIndex);
+	minValueIsNull = PQgetisnull(queryResult, 0, minValueIndex);
+	maxValueIsNull = PQgetisnull(queryResult, 0, maxValueIndex);
 
 	if (!minValueIsNull && !maxValueIsNull)
 	{
-		char *minValueResult = MultiClientGetValue(queryResult, 0, minValueIndex);
-		char *maxValueResult = MultiClientGetValue(queryResult, 0, maxValueIndex);
+		char *minValueResult = PQgetvalue(queryResult, 0, minValueIndex);
+		char *maxValueResult = PQgetvalue(queryResult, 0, maxValueIndex);
 
 		*shardMinValue = cstring_to_text(minValueResult);
 		*shardMaxValue = cstring_to_text(maxValueResult);
 	}
 
-	MultiClientClearResult(queryResult);
-	MultiClientDisconnect(connectionId);
+	PQclear(queryResult);
+	ForgetResults(connection);
 
 	return true;
 }
